@@ -1,58 +1,44 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2012 Fredrik Johansson
 
-******************************************************************************/
+    This file is part of Arb.
+
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
 
 #include "acb.h"
 
 void
-acb_pow_fmpz_binexp(acb_t y, const acb_t b, const fmpz_t e, long prec)
+acb_pow_fmpz_binexp(acb_t y, const acb_t b, const fmpz_t e, slong prec)
 {
-    long i, wp, bits;
+    slong i, wp, bits;
 
-    if (-2L <= *e && *e <= 4L)
+    if (-WORD(2) <= *e && *e <= WORD(4))
     {
-        if (*e == 0L)
+        if (*e == WORD(0))
         {
             acb_one(y);
         }
-        else if (*e == 1L)
+        else if (*e == WORD(1))
         {
             acb_set_round(y, b, prec);
         }
-        else if (*e == -1L)
+        else if (*e == -WORD(1))
         {
             acb_inv(y, b, prec);
         }
-        else if (*e == 2L)
+        else if (*e == WORD(2))
         {
             acb_mul(y, b, b, prec);
         }
-        else if (*e == 3L)
+        else if (*e == WORD(3))
         {
             acb_cube(y, b, prec);
         }
-        else if (*e == 4L)
+        else if (*e == WORD(4))
         {
             acb_mul(y, b, b, prec);
             acb_mul(y, y, y, prec);
@@ -70,8 +56,18 @@ acb_pow_fmpz_binexp(acb_t y, const acb_t b, const fmpz_t e, long prec)
         fmpz_t f;
         fmpz_init(f);
         fmpz_neg(f, e);
-        acb_pow_fmpz_binexp(y, b, f, prec + 2);
-        acb_inv(y, y, prec);
+
+        if (acb_is_exact(b))
+        {
+            acb_pow_fmpz_binexp(y, b, f, prec + 2);
+            acb_inv(y, y, prec);
+        }
+        else
+        {
+            acb_inv(y, b, prec + fmpz_bits(e) + 2);
+            acb_pow_fmpz_binexp(y, y, f, prec);
+        }
+
         fmpz_clear(f);
         return;
     }
@@ -108,13 +104,13 @@ acb_pow_fmpz_binexp(acb_t y, const acb_t b, const fmpz_t e, long prec)
 }
 
 void
-acb_pow_fmpz(acb_t y, const acb_t b, const fmpz_t e, long prec)
+acb_pow_fmpz(acb_t y, const acb_t b, const fmpz_t e, slong prec)
 {
     acb_pow_fmpz_binexp(y, b, e, prec);
 }
 
 void
-acb_pow_ui(acb_t y, const acb_t b, ulong e, long prec)
+acb_pow_ui(acb_t y, const acb_t b, ulong e, slong prec)
 {
     fmpz_t f;
     fmpz_init_set_ui(f, e);
@@ -123,7 +119,7 @@ acb_pow_ui(acb_t y, const acb_t b, ulong e, long prec)
 }
 
 void
-acb_pow_si(acb_t y, const acb_t b, long e, long prec)
+acb_pow_si(acb_t y, const acb_t b, slong e, slong prec)
 {
     fmpz_t f;
     fmpz_init(f);
@@ -133,7 +129,7 @@ acb_pow_si(acb_t y, const acb_t b, long e, long prec)
 }
 
 void
-_acb_pow_exp(acb_t z, const acb_t x, const acb_t y, long prec)
+_acb_pow_exp(acb_t z, const acb_t x, const acb_t y, slong prec)
 {
     acb_t t;
     acb_init(t);
@@ -144,7 +140,7 @@ _acb_pow_exp(acb_t z, const acb_t x, const acb_t y, long prec)
 }
 
 void
-_acb_pow_arb_exp(acb_t z, const acb_t x, const arb_t y, long prec)
+_acb_pow_arb_exp(acb_t z, const acb_t x, const arb_t y, slong prec)
 {
     acb_t t;
     acb_init(t);
@@ -157,7 +153,7 @@ _acb_pow_arb_exp(acb_t z, const acb_t x, const arb_t y, long prec)
 #define BINEXP_LIMIT 64
 
 void
-acb_pow_arb(acb_t z, const acb_t x, const arb_t y, long prec)
+acb_pow_arb(acb_t z, const acb_t x, const arb_t y, slong prec)
 {
     const arf_struct * ymid = arb_midref(y);
     const mag_struct * yrad = arb_radref(y);
@@ -168,6 +164,15 @@ acb_pow_arb(acb_t z, const acb_t x, const arb_t y, long prec)
         return;
     }
 
+    if (acb_is_zero(x))
+    {
+        if (arb_is_positive(y))
+            acb_zero(z);
+        else
+            acb_indeterminate(z);
+        return;
+    }
+
     if (mag_is_zero(yrad))
     {
         /* small half-integer or integer */
@@ -184,17 +189,18 @@ acb_pow_arb(acb_t z, const acb_t x, const arb_t y, long prec)
             }
             else
             {
-                /* hack: give something finite here (should fix sqrt/rsqrt etc) */
-                if (arb_contains_zero(acb_imagref(x)) && arb_contains_nonpositive(acb_realref(x)))
-                {
-                    _acb_pow_arb_exp(z, x, y, prec);
-                    fmpz_clear(e);
-                    return;
-                }
-
                 arf_get_fmpz_fixed_si(e, ymid, -1);
-                acb_sqrt(z, x, prec + fmpz_bits(e));
-                acb_pow_fmpz_binexp(z, z, e, prec);
+                if (fmpz_sgn(e) >= 0)
+                {
+                    acb_sqrt(z, x, prec + fmpz_bits(e));
+                    acb_pow_fmpz_binexp(z, z, e, prec);
+                }
+                else
+                {
+                    fmpz_neg(e, e);
+                    acb_rsqrt(z, x, prec + fmpz_bits(e));
+                    acb_pow_fmpz_binexp(z, z, e, prec);
+                }
             }
 
             fmpz_clear(e);
@@ -206,7 +212,7 @@ acb_pow_arb(acb_t z, const acb_t x, const arb_t y, long prec)
 }
 
 void
-acb_pow(acb_t z, const acb_t x, const acb_t y, long prec)
+acb_pow(acb_t z, const acb_t x, const acb_t y, slong prec)
 {
     if (arb_is_zero(acb_imagref(y)))
     {
@@ -214,6 +220,15 @@ acb_pow(acb_t z, const acb_t x, const acb_t y, long prec)
     }
     else
     {
+        if (acb_is_zero(x))
+        {
+            if (arb_is_positive(acb_realref(y)))
+                acb_zero(z);
+            else
+                acb_indeterminate(z);
+            return;
+        }
+
         _acb_pow_exp(z, x, y, prec);
     }
 }
