@@ -1,27 +1,13 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2013 Fredrik Johansson
 
-******************************************************************************/
+    This file is part of Arb.
+
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
 
 #include "arb.h"
 #include "acb.h"
@@ -33,11 +19,11 @@
 
 #define PI 3.1415926535897932385
 
-static long
-choose_n(double log2z, double argz, int digamma, long prec)
+static slong
+choose_n(double log2z, double argz, int digamma, slong prec)
 {
     double argf, boundn;
-    long n;
+    slong n;
 
     argf = 1.0 / cos(0.5 * argz);
     argf = log(argf) * (1. / log(2));
@@ -56,18 +42,18 @@ choose_n(double log2z, double argz, int digamma, long prec)
         /* if the term magnitude does not decrease, r is too small */
         if (boundn > 1)
         {
-            printf("exception: gamma_stirling_choose_param failed to converge\n");
-            abort();
+            flint_printf("exception: gamma_stirling_choose_param failed to converge\n");
+            flint_abort();
         }
     }
 }
 
 static void
-choose_small(int * reflect, long * r, long * n,
-    double x, double y, int use_reflect, int digamma, long prec)
+choose_small(int * reflect, slong * r, slong * n,
+    double x, double y, int use_reflect, int digamma, slong prec)
 {
     double w, argz, log2z;
-    long rr;
+    slong rr;
 
     /* use reflection formula if very negative */
     if (x < -5.0 && use_reflect)
@@ -98,8 +84,8 @@ choose_small(int * reflect, long * r, long * n,
 }
 
 static void
-choose_large(int * reflect, long * r, long * n,
-    const arf_t a, const arf_t b, int use_reflect, int digamma, long prec)
+choose_large(int * reflect, slong * r, slong * n,
+    const arf_t a, const arf_t b, int use_reflect, int digamma, slong prec)
 {
     if (use_reflect && arf_sgn(a) < 0)
         *reflect = 1;
@@ -109,14 +95,14 @@ choose_large(int * reflect, long * r, long * n,
     *r = 0;
 
     /* so big that we will certainly have n = 0 */
-    if (arf_cmpabs_2exp_si(a, LONG_MAX / 8) >= 0 ||
-        arf_cmpabs_2exp_si(b, LONG_MAX / 8) >= 0)
+    if (arf_cmpabs_2exp_si(a, WORD_MAX / 8) >= 0 ||
+        arf_cmpabs_2exp_si(b, WORD_MAX / 8) >= 0)
     {
         *n = 0;
     }
     else
     {
-        long ab, bb;
+        slong ab, bb;
         double log2z, argz;
 
         ab = arf_abs_bound_lt_2exp_si(a);
@@ -155,8 +141,8 @@ choose_large(int * reflect, long * r, long * n,
 
 
 void
-acb_gamma_stirling_choose_param(int * reflect, long * r, long * n,
-    const acb_t z, int use_reflect, int digamma, long prec)
+acb_gamma_stirling_choose_param(int * reflect, slong * r, slong * n,
+    const acb_t z, int use_reflect, int digamma, slong prec)
 {
     const arf_struct * a = arb_midref(acb_realref(z));
     const arf_struct * b = arb_midref(acb_imagref(z));
@@ -178,8 +164,8 @@ acb_gamma_stirling_choose_param(int * reflect, long * r, long * n,
 }
 
 void
-arb_gamma_stirling_choose_param(int * reflect, long * r, long * n,
-    const arb_t x, int use_reflect, int digamma, long prec)
+arb_gamma_stirling_choose_param(int * reflect, slong * r, slong * n,
+    const arb_t x, int use_reflect, int digamma, slong prec)
 {
     const arf_struct * a = arb_midref(x);
 
@@ -206,7 +192,7 @@ acb_gamma_bound_phase(mag_t bound, const acb_t z)
 {
     arf_t x, y, t, u;
     int xsign;
-    long prec;
+    slong prec;
 
     arf_init(x);
     arf_init(y);
@@ -277,10 +263,10 @@ acb_gamma_bound_phase(mag_t bound, const acb_t z)
   TODO: CHECK n >= 1 ?
 */
 void
-acb_gamma_stirling_bound(mag_ptr err, const acb_t z, long k0, long knum, long n)
+acb_gamma_stirling_bound(mag_ptr err, const acb_t z, slong k0, slong knum, slong n)
 {
     mag_t c, t, u, v;
-    long i, k;
+    slong i, k;
 
     if (arb_contains_zero(acb_imagref(z)) &&
         arb_contains_nonpositive(acb_realref(z)))
@@ -335,7 +321,7 @@ acb_gamma_stirling_bound(mag_ptr err, const acb_t z, long k0, long knum, long n)
 }
 
 void
-arb_gamma_stirling_bound(mag_ptr err, const arb_t x, long k0, long knum, long n)
+arb_gamma_stirling_bound(mag_ptr err, const arb_t x, slong k0, slong knum, slong n)
 {
     acb_t z;
     acb_init(z);
@@ -345,7 +331,7 @@ arb_gamma_stirling_bound(mag_ptr err, const arb_t x, long k0, long knum, long n)
 }
 
 void
-arb_gamma_stirling_coeff(arb_t b, ulong k, int digamma, long prec)
+arb_gamma_stirling_coeff(arb_t b, ulong k, int digamma, slong prec)
 {
     fmpz_t d;
     fmpz_init(d);
@@ -364,12 +350,12 @@ arb_gamma_stirling_coeff(arb_t b, ulong k, int digamma, long prec)
 }
 
 void
-arb_gamma_stirling_eval(arb_t s, const arb_t z, long nterms, int digamma, long prec)
+arb_gamma_stirling_eval(arb_t s, const arb_t z, slong nterms, int digamma, slong prec)
 {
     arb_t b, t, logz, zinv, zinv2;
     mag_t err;
 
-    long k, term_prec;
+    slong k, term_prec;
     double z_mag, term_mag;
 
     arb_init(b);
@@ -451,10 +437,10 @@ arb_gamma_stirling_eval(arb_t s, const arb_t z, long nterms, int digamma, long p
 }
 
 void
-arb_gamma_fmpq_stirling(arb_t y, const fmpq_t a, long prec)
+arb_gamma_fmpq_stirling(arb_t y, const fmpq_t a, slong prec)
 {
     int reflect;
-    long r, n, wp;
+    slong r, n, wp;
     arb_t x, t, u, v;
 
     wp = prec + FLINT_BIT_COUNT(prec);
@@ -504,11 +490,11 @@ arb_gamma_fmpq_stirling(arb_t y, const fmpq_t a, long prec)
 }
 
 void
-arb_gamma_const_1_3_eval(arb_t s, long prec)
+arb_gamma_const_1_3_eval(arb_t s, slong prec)
 {
     hypgeom_t series;
     arb_t t, u;
-    long wp = prec + 4 + 2 * FLINT_BIT_COUNT(prec);
+    slong wp = prec + 4 + 2 * FLINT_BIT_COUNT(prec);
 
     arb_init(t);
     arb_init(u);
@@ -533,8 +519,8 @@ arb_gamma_const_1_3_eval(arb_t s, long prec)
     arb_mul(s, s, u, wp);
 
     arb_div(s, s, t, wp);
-    arb_root(s, s, 2, wp);
-    arb_root(s, s, 3, prec);
+    arb_root_ui(s, s, 2, wp);
+    arb_root_ui(s, s, 3, prec);
 
     hypgeom_clear(series);
     arb_clear(t);
@@ -544,10 +530,10 @@ arb_gamma_const_1_3_eval(arb_t s, long prec)
 ARB_DEF_CACHED_CONSTANT(arb_gamma_const_1_3, arb_gamma_const_1_3_eval)
 
 void
-arb_gamma_const_1_4_eval(arb_t x, long prec)
+arb_gamma_const_1_4_eval(arb_t x, slong prec)
 {
     arb_t t, u;
-    long wp = prec + 4 + 2 * FLINT_BIT_COUNT(prec);
+    slong wp = prec + 4 + 2 * FLINT_BIT_COUNT(prec);
 
     arb_init(t);
     arb_init(u);
@@ -571,9 +557,9 @@ arb_gamma_const_1_4_eval(arb_t x, long prec)
 ARB_DEF_CACHED_CONSTANT(arb_gamma_const_1_4, arb_gamma_const_1_4_eval)
 
 void
-arb_gamma_small_frac(arb_t y, unsigned int p, unsigned int q, long prec)
+arb_gamma_small_frac(arb_t y, unsigned int p, unsigned int q, slong prec)
 {
-    long wp = prec + 4;
+    slong wp = prec + 4;
 
     if (q == 1)
     {
@@ -628,7 +614,7 @@ arb_gamma_small_frac(arb_t y, unsigned int p, unsigned int q, long prec)
         arb_div_ui(t, t, 3, wp);
         arb_sqrt(t, t, wp);
         arb_set_ui(y, 2);
-        arb_root(y, y, 3, wp);
+        arb_root_ui(y, y, 3, wp);
         arb_mul(t, t, y, wp);
         arb_gamma_const_1_3(y, wp);
         arb_mul(y, y, y, prec);
@@ -649,18 +635,18 @@ arb_gamma_small_frac(arb_t y, unsigned int p, unsigned int q, long prec)
     }
     else
     {
-        printf("small fraction not implemented!\n");
-        abort();
+        flint_printf("small fraction not implemented!\n");
+        flint_abort();
     }
 }
 
 void
-arb_gamma_fmpq_outward(arb_t y, const fmpq_t x, long prec)
+arb_gamma_fmpq_outward(arb_t y, const fmpq_t x, slong prec)
 {
     fmpq_t a;
     fmpz_t n;
     fmpz p, q;
-    long m;
+    slong m;
     arb_t t, u;
 
     fmpq_init(a);
@@ -682,8 +668,8 @@ arb_gamma_fmpq_outward(arb_t y, const fmpq_t x, long prec)
 
     if (!fmpz_fits_si(n))
     {
-        printf("gamma: too large fmpq to reduce to 0!\n");
-        abort();
+        flint_printf("gamma: too large fmpq to reduce to 0!\n");
+        flint_abort();
     }
 
     m = fmpz_get_si(n);
@@ -698,8 +684,8 @@ arb_gamma_fmpq_outward(arb_t y, const fmpq_t x, long prec)
     }
     else
     {
-        printf("arb_gamma_fmpq: invalid fraction\n");
-        abort();
+        flint_printf("arb_gamma_fmpq: invalid fraction\n");
+        flint_abort();
     }
 
     /* argument reduction */
@@ -721,7 +707,7 @@ arb_gamma_fmpq_outward(arb_t y, const fmpq_t x, long prec)
 }
 
 void
-arb_gamma_fmpq(arb_t y, const fmpq_t x, long prec)
+arb_gamma_fmpq(arb_t y, const fmpq_t x, slong prec)
 {
     fmpz p, q;
 
@@ -762,20 +748,21 @@ arb_gamma_fmpq(arb_t y, const fmpq_t x, long prec)
 }
 
 void
-arb_gamma_fmpz(arb_t y, const fmpz_t x, long prec)
+arb_gamma_fmpz(arb_t y, const fmpz_t x, slong prec)
 {
     fmpq_t t;
     *fmpq_numref(t) = *x;
-    *fmpq_denref(t) = 1L;
+    *fmpq_denref(t) = WORD(1);
     arb_gamma_fmpq(y, t, prec);
 }
 
 static void
-_arb_gamma(arb_t y, const arb_t x, long prec, int inverse)
+_arb_gamma(arb_t y, const arb_t x, slong prec, int inverse)
 {
     int reflect;
-    long r, n, wp;
+    slong r, n, wp;
     arb_t t, u, v;
+    double acc;
 
     if (arb_is_exact(x))
     {
@@ -813,7 +800,12 @@ _arb_gamma(arb_t y, const arb_t x, long prec, int inverse)
         }
     }
 
-    wp = prec + FLINT_BIT_COUNT(prec);
+    /* todo: for large x (if exact or accurate enough), increase precision */
+    acc = arb_rel_accuracy_bits(x);
+    acc = FLINT_MAX(acc, 0);
+    wp = FLINT_MIN(prec, acc + 20);
+    wp = FLINT_MAX(wp, 2);
+    wp = wp + FLINT_BIT_COUNT(wp);
 
     arb_gamma_stirling_choose_param(&reflect, &r, &n, x, 1, 0, wp);
 
@@ -855,24 +847,43 @@ _arb_gamma(arb_t y, const arb_t x, long prec, int inverse)
 }
 
 void
-arb_gamma(arb_t y, const arb_t x, long prec)
+arb_gamma(arb_t y, const arb_t x, slong prec)
 {
     _arb_gamma(y, x, prec, 0);
 }
 
 void
-arb_rgamma(arb_t y, const arb_t x, long prec)
+arb_rgamma(arb_t y, const arb_t x, slong prec)
 {
     _arb_gamma(y, x, prec, 1);
 }
 
 void
-arb_lgamma(arb_t y, const arb_t x, long prec)
+arb_lgamma(arb_t y, const arb_t x, slong prec)
 {
     int reflect;
-    long r, n, wp;
+    slong r, n, wp;
     arb_t t, u;
 
+    if (!arb_is_positive(x))
+    {
+        arb_indeterminate(y);
+        return;
+    }
+
+    /* fast gamma(n), gamma(n/2) or gamma(n/4) */
+    if (arb_is_exact(x) && arf_is_int_2exp_si(arb_midref(x), -2) &&
+            arf_cmpabs_2exp_si(arb_midref(x), prec) < 0)
+    {
+        fmpq_t a;
+        fmpq_init(a);
+        arf_get_fmpq(a, arb_midref(x));
+        arb_gamma_fmpq(y, a, prec);
+        arb_log(y, y, prec);
+        fmpq_clear(a);
+        return;
+    }
+
     wp = prec + FLINT_BIT_COUNT(prec);
 
     arb_gamma_stirling_choose_param(&reflect, &r, &n, x, 0, 0, wp);
