@@ -1,189 +1,25 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2012-2014 Fredrik Johansson
 
-******************************************************************************/
+    This file is part of Arb.
 
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
+
+#include "flint/mpn_extras.h"
 #include "arb.h"
-#include "mpn_extras.h"
 
 #define TMP_ALLOC_LIMBS(__n) TMP_ALLOC((__n) * sizeof(mp_limb_t))
-
-int _arf_get_integer_mpn(mp_ptr y, mp_srcptr x, mp_size_t xn, long exp);
-
-int _arf_set_mpn_fixed(arf_t z, mp_srcptr xp, mp_size_t xn, mp_size_t fixn, int negative, long prec);
-
 #define MAGLIM(prec) FLINT_MAX(65536, (4*prec))
 
-static void
-_arf_sin(arf_t z, const arf_t x, long prec, arf_rnd_t rnd)
-{
-    mpfr_t xf, zf;
-    mp_ptr zptr, tmp;
-    mp_srcptr xptr;
-    mp_size_t xn, zn, val;
-    TMP_INIT;
-    TMP_START;
-
-    zn = (prec + FLINT_BITS - 1) / FLINT_BITS;
-    tmp = TMP_ALLOC(zn * sizeof(mp_limb_t));
-
-    ARF_GET_MPN_READONLY(xptr, xn, x);
-
-    xf->_mpfr_d = (mp_ptr) xptr;
-    xf->_mpfr_prec = xn * FLINT_BITS;
-    xf->_mpfr_sign = ARF_SGNBIT(x) ? -1 : 1;
-    xf->_mpfr_exp = ARF_EXP(x);
-
-    zf->_mpfr_d = tmp;
-    zf->_mpfr_prec = prec;
-    zf->_mpfr_sign = 1;
-    zf->_mpfr_exp = 0;
-
-    mpfr_set_emin(MPFR_EMIN_MIN);
-    mpfr_set_emax(MPFR_EMAX_MAX);
-
-    mpfr_sin(zf, xf, arf_rnd_to_mpfr(rnd));
-
-    val = 0;
-    while (tmp[val] == 0)
-        val++;
-
-    ARF_GET_MPN_WRITE(zptr, zn - val, z);
-    flint_mpn_copyi(zptr, tmp + val, zn - val);
-    if (zf->_mpfr_sign < 0)
-        ARF_NEG(z);
-
-    fmpz_set_si(ARF_EXPREF(z), zf->_mpfr_exp);
-
-    TMP_END;
-}
-
-static void
-_arf_cos(arf_t z, const arf_t x, long prec, arf_rnd_t rnd)
-{
-    mpfr_t xf, zf;
-    mp_ptr zptr, tmp;
-    mp_srcptr xptr;
-    mp_size_t xn, zn, val;
-    TMP_INIT;
-    TMP_START;
-
-    zn = (prec + FLINT_BITS - 1) / FLINT_BITS;
-    tmp = TMP_ALLOC(zn * sizeof(mp_limb_t));
-
-    ARF_GET_MPN_READONLY(xptr, xn, x);
-
-    xf->_mpfr_d = (mp_ptr) xptr;
-    xf->_mpfr_prec = xn * FLINT_BITS;
-    xf->_mpfr_sign = ARF_SGNBIT(x) ? -1 : 1;
-    xf->_mpfr_exp = ARF_EXP(x);
-
-    zf->_mpfr_d = tmp;
-    zf->_mpfr_prec = prec;
-    zf->_mpfr_sign = 1;
-    zf->_mpfr_exp = 0;
-
-    mpfr_set_emin(MPFR_EMIN_MIN);
-    mpfr_set_emax(MPFR_EMAX_MAX);
-
-    mpfr_cos(zf, xf, arf_rnd_to_mpfr(rnd));
-
-    val = 0;
-    while (tmp[val] == 0)
-        val++;
-
-    ARF_GET_MPN_WRITE(zptr, zn - val, z);
-    flint_mpn_copyi(zptr, tmp + val, zn - val);
-    if (zf->_mpfr_sign < 0)
-        ARF_NEG(z);
-
-    fmpz_set_si(ARF_EXPREF(z), zf->_mpfr_exp);
-
-    TMP_END;
-}
-
-static void
-_arf_sin_cos(arf_t z, arf_t w, const arf_t x, long prec, arf_rnd_t rnd)
-{
-    mpfr_t xf, zf, wf;
-    mp_ptr zptr, wptr, tmp, tmp2;
-    mp_srcptr xptr;
-    mp_size_t xn, zn, wn, val;
-    TMP_INIT;
-    TMP_START;
-
-    zn = wn = (prec + FLINT_BITS - 1) / FLINT_BITS;
-    tmp = TMP_ALLOC(2 * zn * sizeof(mp_limb_t));
-    tmp2 = tmp + zn;
-
-    ARF_GET_MPN_READONLY(xptr, xn, x);
-
-    xf->_mpfr_d = (mp_ptr) xptr;
-    xf->_mpfr_prec = xn * FLINT_BITS;
-    xf->_mpfr_sign = ARF_SGNBIT(x) ? -1 : 1;
-    xf->_mpfr_exp = ARF_EXP(x);
-
-    zf->_mpfr_d = tmp;
-    zf->_mpfr_prec = prec;
-    zf->_mpfr_sign = 1;
-    zf->_mpfr_exp = 0;
-
-    wf->_mpfr_d = tmp2;
-    wf->_mpfr_prec = prec;
-    wf->_mpfr_sign = 1;
-    wf->_mpfr_exp = 0;
-
-    mpfr_set_emin(MPFR_EMIN_MIN);
-    mpfr_set_emax(MPFR_EMAX_MAX);
-
-    mpfr_sin_cos(zf, wf, xf, arf_rnd_to_mpfr(rnd));
-
-    val = 0;
-    while (tmp[val] == 0)
-        val++;
-    ARF_GET_MPN_WRITE(zptr, zn - val, z);
-    flint_mpn_copyi(zptr, tmp + val, zn - val);
-    if (zf->_mpfr_sign < 0)
-        ARF_NEG(z);
-    fmpz_set_si(ARF_EXPREF(z), zf->_mpfr_exp);
-
-    val = 0;
-    while (tmp2[val] == 0)
-        val++;
-    ARF_GET_MPN_WRITE(wptr, wn - val, w);
-    flint_mpn_copyi(wptr, tmp2 + val, wn - val);
-    if (wf->_mpfr_sign < 0)
-        ARF_NEG(w);
-    fmpz_set_si(ARF_EXPREF(w), wf->_mpfr_exp);
-
-    TMP_END;
-}
-
 void
-arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
+arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, slong prec)
 {
     int want_sin, want_cos;
-    long exp, wp, wn, N, r, wprounded;
+    slong exp, wp, wn, N, r, wprounded;
     mp_ptr tmp, w, sina, cosa, sinb, cosb, ta, tb;
     mp_ptr sinptr, cosptr;
     mp_limb_t p1, q1bits, p2, q2bits, error, error2;
@@ -212,22 +48,7 @@ arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
     /* Too high precision to use table -- use generic algorithm */
     if (wp > ARB_SIN_COS_TAB2_PREC)
     {
-        if (want_sin && want_cos)
-        {
-            _arf_sin_cos(arb_midref(zsin), arb_midref(zcos), x, prec, ARB_RND);
-            arf_mag_set_ulp(arb_radref(zsin), arb_midref(zsin), prec);
-            arf_mag_set_ulp(arb_radref(zcos), arb_midref(zcos), prec);
-        }
-        else if (want_sin)
-        {
-            _arf_sin(arb_midref(zsin), x, prec, ARB_RND);
-            arf_mag_set_ulp(arb_radref(zsin), arb_midref(zsin), prec);
-        }
-        else
-        {
-            _arf_cos(arb_midref(zcos), x, prec, ARB_RND);
-            arf_mag_set_ulp(arb_radref(zcos), arb_midref(zcos), prec);
-        }
+        arb_sin_cos_arf_generic(zsin, zcos, x, prec);
         return;
     }
 
@@ -245,22 +66,7 @@ arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
     if (_arb_get_mpn_fixed_mod_pi4(w, NULL, &octant, &error, x, wn) == 0)
     {
         /* may run out of precision for pi/4 */
-        if (want_sin && want_cos)
-        {
-            _arf_sin_cos(arb_midref(zsin), arb_midref(zcos), x, prec, ARB_RND);
-            arf_mag_set_ulp(arb_radref(zsin), arb_midref(zsin), prec);
-            arf_mag_set_ulp(arb_radref(zcos), arb_midref(zcos), prec);
-        }
-        else if (want_sin)
-        {
-            _arf_sin(arb_midref(zsin), x, prec, ARB_RND);
-            arf_mag_set_ulp(arb_radref(zsin), arb_midref(zsin), prec);
-        }
-        else
-        {
-            _arf_cos(arb_midref(zcos), x, prec, ARB_RND);
-            arf_mag_set_ulp(arb_radref(zcos), arb_midref(zcos), prec);
-        }
+        arb_sin_cos_arf_generic(zsin, zcos, x, prec);
         TMP_END;
         return;
     }
@@ -307,14 +113,14 @@ arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
         /* Taylor series evaluation error */
         error += error2;
         /* Taylor series truncation error */
-        error += 1UL << (wprounded-wp);
+        error += UWORD(1) << (wprounded-wp);
     }
     else  /* Compute cos(a) from sin(a) using a square root. */
     {
         /* Evaluate Taylor series */
         _arb_sin_cos_taylor_rs(sina, cosa, &error2, w, wn, N, 1, 1);
         error += error2;
-        error += 1UL << (wprounded-wp);
+        error += UWORD(1) << (wprounded-wp);
 
         if (flint_mpn_zero_p(sina, wn))
         {
@@ -462,7 +268,7 @@ arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
     if (want_sin)
     {
         inexact = _arf_set_mpn_fixed(arb_midref(zsin), sinptr,
-            wn, wn, sinnegative, prec);
+            wn, wn, sinnegative, prec, ARB_RND);
         if (inexact)
             arf_mag_add_ulp(arb_radref(zsin),
                 arb_radref(zsin), arb_midref(zsin), prec);
@@ -471,7 +277,7 @@ arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
     if (want_cos)
     {
         inexact = _arf_set_mpn_fixed(arb_midref(zcos), cosptr,
-            wn, wn, cosnegative, prec);
+            wn, wn, cosnegative, prec, ARB_RND);
         if (inexact)
             arf_mag_add_ulp(arb_radref(zcos),
                 arb_radref(zcos), arb_midref(zcos), prec);
@@ -482,7 +288,7 @@ arb_sin_cos_arf_new(arb_t zsin, arb_t zcos, const arf_t x, long prec)
 
 
 void
-arb_sin_arf(arb_t s, const arf_t x, long prec, long maglim)
+arb_sin_arf(arb_t s, const arf_t x, slong prec, slong maglim)
 {
     if (arf_is_special(x))
     {
@@ -502,7 +308,7 @@ arb_sin_arf(arb_t s, const arf_t x, long prec, long maglim)
     }
     else
     {
-        long xmag;
+        slong xmag;
 
         /* 2^(xmag-1) <= |x| < 2^xmag */
         xmag = ARF_EXP(x);
@@ -532,7 +338,7 @@ arb_sin_arf(arb_t s, const arf_t x, long prec, long maglim)
 }
 
 void
-arb_cos_arf(arb_t c, const arf_t x, long prec, long maglim)
+arb_cos_arf(arb_t c, const arf_t x, slong prec, slong maglim)
 {
     if (arf_is_special(x))
     {
@@ -552,7 +358,7 @@ arb_cos_arf(arb_t c, const arf_t x, long prec, long maglim)
     }
     else
     {
-        long xmag;
+        slong xmag;
 
         /* 2^(xmag-1) <= |x| < 2^xmag */
         xmag = ARF_EXP(x);
@@ -581,7 +387,7 @@ arb_cos_arf(arb_t c, const arf_t x, long prec, long maglim)
 }
 
 void
-arb_sin_cos_arf(arb_t s, arb_t c, const arf_t x, long prec, long maglim)
+arb_sin_cos_arf(arb_t s, arb_t c, const arf_t x, slong prec, slong maglim)
 {
     if (arf_is_special(x))
     {
@@ -604,7 +410,7 @@ arb_sin_cos_arf(arb_t s, arb_t c, const arf_t x, long prec, long maglim)
     }
     else
     {
-        long xmag;
+        slong xmag;
 
         /* 2^(xmag-1) <= |x| < 2^xmag */
         xmag = ARF_EXP(x);
@@ -640,7 +446,7 @@ arb_sin_cos_arf(arb_t s, arb_t c, const arf_t x, long prec, long maglim)
 }
 
 void
-arb_sin(arb_t s, const arb_t x, long prec)
+arb_sin(arb_t s, const arb_t x, slong prec)
 {
     if (arb_is_exact(x))
     {
@@ -664,7 +470,7 @@ arb_sin(arb_t s, const arb_t x, long prec)
 }
 
 void
-arb_cos(arb_t s, const arb_t x, long prec)
+arb_cos(arb_t s, const arb_t x, slong prec)
 {
     if (arb_is_exact(x))
     {
@@ -688,7 +494,7 @@ arb_cos(arb_t s, const arb_t x, long prec)
 }
 
 void
-arb_sin_cos(arb_t s, arb_t c, const arb_t x, long prec)
+arb_sin_cos(arb_t s, arb_t c, const arb_t x, slong prec)
 {
     if (arb_is_exact(x))
     {
